<?php

namespace GGM\ProductManagerImplementation\ProductManagerMagento2;

use Magento;
use Magento\Catalog\Api\ProductRepositoryInterface;
use Magento\Catalog\Model\Product;
use Magento\Catalog\Model\Product\Visibility;
use Magento\Catalog\Model\ProductFactory;
use Magento\CatalogUrlRewrite\Model\ProductUrlPathGenerator;
use Magento\CatalogUrlRewrite\Model\ProductUrlRewriteGenerator;
use Magento\Framework\Api\SearchCriteriaBuilder;
use Magento\Framework\App\CacheInterface;
use Magento\Framework\App\ProductMetadataInterface;
use Magento\Indexer\Model\Indexer\CollectionFactory;
use Magento\Store\Model\StoreManager;
use Magento\UrlRewrite\Model\UrlFinderInterface;
use Magento\UrlRewrite\Model\UrlPersistInterface;
use Magento\UrlRewrite\Service\V1\Data\UrlRewrite;

class ProductManagerReindexHelper
{
    protected $_storeManager;
    protected $_productModel;
    protected $_productFactory;
    protected $_productRepository;
    protected $_indexerCollectionFactory;
    protected $_productCollectionFactory;
    protected $_urlRewriteGenerator;
    protected $_urlRewrite;
    protected $_urlPersist;
    protected $_urlFinder;
    protected $_cacheManager;
    protected $_productMetadata;
    protected $_searchCriteriaBuilder;
    protected $_productUrlPathGenerator;

    public function __construct(
        StoreManager $storeManager,
        Product $productModel,
        ProductFactory $productFactory,
        ProductRepositoryInterface $productRepository,
        CollectionFactory $indexerCollectionFactory,
        Magento\Catalog\Model\ResourceModel\Product\CollectionFactory $productCollectionFactory,
        ProductUrlRewriteGenerator $urlRewriteGenerator,
        UrlRewrite $urlRewrite,
        UrlPersistInterface $urlPersist,
        UrlFinderInterface $urlFinder,
        ProductUrlPathGenerator $productUrlPathGenerator,
        CacheInterface $cacheManager,
        ProductMetadataInterface $productMetadata,
        SearchCriteriaBuilder $searchCriteriaBuilder
    ) {
        $this->_storeManager = $storeManager;
        $this->_productModel = $productModel;
        $this->_productFactory = $productFactory;
        $this->_productRepository = $productRepository;
        $this->_indexerCollectionFactory = $indexerCollectionFactory;
        $this->_productCollectionFactory = $productCollectionFactory;
        $this->_urlRewriteGenerator = $urlRewriteGenerator;
        $this->_urlRewrite = $urlRewrite;
        $this->_urlPersist = $urlPersist;
        $this->_urlFinder = $urlFinder;
        $this->_productUrlPathGenerator = $productUrlPathGenerator;
        $this->_cacheManager = $cacheManager;
        $this->_productMetadata = $productMetadata;
        $this->_searchCriteriaBuilder = $searchCriteriaBuilder;
    }

    public function loadAndSaveProducts(&$result, $productIds)
    {
        $this->_storeManager->setCurrentStore('admin');

        $count = count($productIds);

        for ($i = 0; $i < $count; ++$i) {
            $productId = $productIds[$i];

            $startTime = microtime(true);

            $r = new \StdClass();
            $r->productId = $productId;

            try {
                $product = $this->_productRepository->getById($productId);

                if ($product) {
                    $product->setMediaGalleryEntries($product->getMediaGalleryEntries());

                    if (!$this->_productRepository->save($product)) {
                        $r->error = "Load and save product error (product $productId):" . " save failed product";
                    } else {
                        $r->result = "Load and save product successful (product $productId)";
                    }
                }
            } catch (\Exception $e) {
                $r->error = "Load and save product error (product $productId):" . $e->getMessage() . '   stack trace: ' . $e->getTraceAsString();
            }
            $this->_cacheManager->clean('catalog_product_' . $productId);
        }

        $r->executionTime = microtime(true) - $startTime;
        $result = $r;
    }

    public function regenerateProductsUrlRewrites(&$result, $productIds)
    {
        $this->_storeManager->setCurrentStore('admin');

        $stores = $this->_storeManager->getStores(false);
        $a = array();

        foreach ($stores as $store) {

            $collection = $this->_productCollectionFactory->create();
            $storeId = $store->getId();

            $collection->addStoreFilter($store->getId())
                ->setStoreId($store->getId());

            if (!empty($productIds)) {
                $collection->addIdFilter($productIds);
            }

            $collection->addAttributeToSelect(['url_path', 'url_key', 'visibility']);
            $productList = $collection->load();

            foreach ($productList as $product) {
                $startTime = microtime(true);
                $r = new \StdClass();
                $r->productId = $product->getId();
                $r->storeId = $store->getId();
                $r->storeCode = $store->getCode();
                $r->urlKey = $product->getUrlKey();
                $r->visibility = $product->getVisibility();

                switch ($product->getVisibility()) {
                    case Visibility::VISIBILITY_NOT_VISIBLE:
                        $r->visibilityString = "Not Visible Individually";
                    break;
                    case Visibility::VISIBILITY_IN_CATALOG:
                        $r->visibilityString = "Catalog";
                    break;
                    case Visibility::VISIBILITY_IN_SEARCH:
                        $r->visibilityString = "Search";
                    break;
                    case Visibility::VISIBILITY_BOTH:
                        $r->visibilityString = "Catalog, Search";
                    break;
                }
                $existingUrls = $this->_urlFinder->findAllByData([
                    UrlRewrite::ENTITY_ID => $product->getId(),
                    UrlRewrite::ENTITY_TYPE => ProductUrlRewriteGenerator::ENTITY_TYPE,
                    UrlRewrite::REDIRECT_TYPE => 0,
                    UrlRewrite::STORE_ID => $store->getId(),
                ]);

                $r->existingUrlRewrites = array();
                foreach ($existingUrls as &$urlRewrite) {
                    $u = new \StdClass();
                    $u->url_rewrite_id = $urlRewrite->getUrlRewriteId();
                    $u->entity_type = $urlRewrite->getEntityType();
                    $u->entity_id = $urlRewrite->getEntityId();
                    $u->request_path = $urlRewrite->getRequestPath();
                    $u->target_path = $urlRewrite->getTargetPath();
                    $u->redirect_type = $urlRewrite->getRedirectType();
                    $u->storeId = $urlRewrite->getStoreId();
                    $r->existingUrlRewrites[] = $u;
                }

                if ($product->isVisibleInSiteVisibility()) {
                    try {
                        $product->unsUrlPath();
                        $urlPath = $this->_productUrlPathGenerator->getUrlPath($product);
                        $product->setUrlPath($urlPath);
                        $newUrls = $this->_urlRewriteGenerator->generate($product);

                        $r->newUrlRewrites = array();
                        foreach ($newUrls as &$urlRewrite) {
                            $u = new \StdClass();
                            $u->entity_type = $urlRewrite->getEntityType();
                            $u->entity_id = $urlRewrite->getEntityId();
                            $u->request_path = $urlRewrite->getRequestPath();
                            $u->target_path = $urlRewrite->getTargetPath();
                            $u->storeId = $urlRewrite->getStoreId();
                            $r->newUrlRewrites[] = $u;
                        }

                        if (!$this->compareUrlRewriteArrays($newUrls, $existingUrls)) {
                            foreach ($newUrls as $newUrl) {
                                $this->_urlPersist->deleteByData([
                                    UrlRewrite::REQUEST_PATH => $newUrl->getRequestPath(),
                                    UrlRewrite::REDIRECT_TYPE => 301,
                                    UrlRewrite::STORE_ID => $store->getId()
                                ]);
                            }
                            $this->_urlPersist->replace($newUrls);
                            $r->result = 'Url rewrites replaced';
                        } else {
                            $r->result = 'Url rewrites unchanged';
                        }
                    } catch (\Exception $e) {
                        $r->error = 'Url rewrites error';
                        $r->message = $e->getMessage();
                    }

                    $r->executionTime = microtime(true) - $startTime;
                    $a[] = $r;
                } else {
                    $this->_urlPersist->deleteByData([
                        UrlRewrite::ENTITY_ID => $product->getId(),
                        UrlRewrite::ENTITY_TYPE => ProductUrlRewriteGenerator::ENTITY_TYPE,
                        UrlRewrite::REDIRECT_TYPE => 0,
                        UrlRewrite::STORE_ID => $store->getId()
                    ]);

                    $r->result = 'Url rewrites deleted';
                    $r->comment = 'Existing URLS deleted because product is not visible';
                    $a[] = $r;
                }
            }
        }

        $result->result = $a;
    }


    public function simplifyUrlRewriteArray(array $urls)
    {
        $a = array();

        foreach ($urls as $url) {
            if ($url->getRedirectType() == 0) {
                $a[] = [$url->getRequestPath(), $url->getTargetPath()];
            }
        }

        asort($a);

        return $a;
    }


    public function compareUrlRewriteArrays(array $newUrls, array $existingUrls)
    {
        $a = $this->simplifyUrlRewriteArray($newUrls);
        $b = $this->simplifyUrlRewriteArray($existingUrls);
        $result = ($a == $b);
        return $result;
    }


    public function reindexProductsUsingIndexers(&$jsonRpcResult, $productIds, $reindexerIds)
    {
        $skipIndexerIds = ['design_config_grid', 'customer_grid', 'elasticsuite_thesaurus'];
        $a = array();

        $this->_storeManager->setCurrentStore('admin');

        $indexerCollection = $this->_indexerCollectionFactory->create();
        $indexerIds = $indexerCollection->getAllIds();

        $productsIdsString = implode(',', $productIds);

        foreach ($indexerCollection->getItems() as $indexer) {
            $indexerId = $indexer->getId();
            $skipIndexer = false;

            if (in_array($indexerId, $skipIndexerIds)) {
                $skipIndexer = true;
            }

            if (!empty($reindexerIds) && !in_array($indexerId, $reindexerIds)) {
                $skipIndexer = true;
            }

            if ($skipIndexer) {
                continue;
            }

            $startTime = microtime(true);
            if ($indexerId == "inventory") {
                $skus = $this->getSkus($productIds);
                $sourceItems = $this->getSourceItems($skus);
                if (method_exists($indexer, 'executeList')) {
                    $indexer->executeList($sourceItems);
                } else {
                    $indexer->reindexList($sourceItems);
                }
            } else {
                if (method_exists($indexer, 'executeList')) {
                    $indexer->executeList($productIds);
                } else {
                    $indexer->reindexList($productIds);
                }
            }

            $r = new \StdClass();
            $r->result = "Reindex $indexerId successful (product $productsIdsString)";
            $r->executionTime = microtime(true) - $startTime;
            $a[] = $r;
        }

        $jsonRpcResult->result = $a;
    }

    public function getSkus($productIds)
    {
        $collection = $this->_productCollectionFactory->create();

        if (!empty($productIds)) {
            $collection->addIdFilter($productIds);
        }

        $collection->addAttributeToSelect(['sku']);
        $productList = $collection->load();

        $skus = array();
        foreach ($collection as $product) {
            $skus[] = $product->getSku();
        }

        return $skus;
    }

    public function getSourceItems($skus)
    {
        $sourceItems = array();

        return $sourceItems;
    }


    public function cleanProductsCache(&$result, $productIds)
    {
        $this->_storeManager->setCurrentStore('admin');

        $count = count($productIds);

        for ($i = 0; $i < $count; ++$i) {
            $productId = $productIds[$i];
            $this->_cacheManager->clean('catalog_product_' . $productId);
        }

        $productsIdsString = implode(',', $productIds);
        $result->result = "product cache cleaned (product $productsIdsString)";
    }
}